from __future__ import annotations

import logging
from typing import Any, Dict

from actionweaver.actions import orchestration
from actionweaver.actions.orchestration import Orchestration
from actionweaver.actions.orchestration_expr import (
    RequireNext,
    SelectOne,
    _ActionHandlerLLMInvoke,
    _ActionHandlerRequired,
    _ActionHandlerSelectOne,
)
from actionweaver.utils import DEFAULT_ACTION_SCOPE
from actionweaver.utils.pydantic_utils import create_pydantic_model_from_func


class ActionException(Exception):
    pass


def action(
    name,
    scope=DEFAULT_ACTION_SCOPE,
    logger=None,
    orch_expr=None,
    models=[],
    stop=False,
):
    """
    Decorator function to create an Action object.

    Args:
    - name (str): Name of the action.
    - scope (str): Scope of the action, default is DEFAULT_ACTION_SCOPE.
    - logger (logging.Logger): Logger instance to log information, default is None.
    - models (list[pydantic.BaseModel]): List of pydantic models to be used in the action.
    - stop (bool): If True, the agent will stop immediately after invoking this action.
    Returns:
    - create_action: A function that takes a decorated object and returns an Action object.
    """

    _logger = logger or logging.getLogger(__name__)

    def create_action(decorated_obj):
        _logger.debug({"message": f"Creating action with name: {name}, scope: {scope}"})

        action = Action(
            name=name,
            scope=scope,
            decorated_obj=decorated_obj,
            orch_expr=orch_expr,
            logger=_logger,
            stop=stop,
        ).build_pydantic_model_cls(models=models)

        return action

    return create_action


class Action:
    def __init__(
        self,
        name,
        decorated_obj,
        scope=None,
        orch_expr=None,
        logger=None,
        stop=False,
    ):
        self.name = name
        self.scope = scope or DEFAULT_ACTION_SCOPE
        self.logger = logger
        self.orch_expr = orch_expr
        self.stop = stop

        if decorated_obj.__doc__ is None:
            raise ActionException(
                f"Decorated method under action {name} must have a docstring for description."
            )

        self.description = decorated_obj.__doc__
        self.pydantic_cls = None

        self.decorated_method = decorated_obj

        self.__module__ = self.decorated_method.__module__
        self.__name__ = self.decorated_method.__name__
        self.__qualname__ = self.decorated_method.__qualname__
        self.__annotations__ = self.decorated_method.__annotations__
        self.__doc__ = self.decorated_method.__doc__

    def build_pydantic_model_cls(
        self,
        models=None,
        override_params=None,  # override_params: Optional dictionary of parameters to override kwarg and non-kwarg of decorated method.
    ):
        if models is None:
            models = []

        self.pydantic_cls = create_pydantic_model_from_func(
            self.decorated_method,
            self.decorated_method.__name__.title(),
            models=models,
            override_params=override_params,
        )
        return self

    def json_schema(self):
        return self.pydantic_cls.model_json_schema()

    def invoke(self, chat, messages, force=True, stream=False):
        assert len(messages) > 0, "Messages cannot be empty"

        if messages is None:
            messages = []

        return chat.create(
            messages,
            actions=[self],
            orch_expr=RequireNext([self.name]) if force else None,
            stream=stream,
        )

    def bind(self, instance) -> InstanceAction:
        return InstanceAction(
            self.name,
            self.decorated_method,
            self.pydantic_cls,
            self.scope,
            self.orch_expr,
            self.logger,
            self.stop,
            instance=instance,
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self.logger:
            self.logger.debug(
                {
                    "message": f"[Action {self.name}, method {self.__name__}] Calling action: {self.name} with args: {args}"
                }
            )
        response = self.decorated_method(*args, **kwargs)
        if self.logger:
            self.logger.debug(
                {
                    "message": f"[Action {self.name}, method {self.__name__}] Received response: {response}"
                }
            )
        return response

    def __get__(self, instance, owner) -> InstanceAction:
        """

        Note:
            The `__get__` method is a descriptor method that is called when the action is accessed from an instance.
            It returns an instance-specific action method that is bound to the given instance.
        """
        return InstanceAction(
            self.name,
            self.decorated_method,
            self.pydantic_cls,
            self.scope,
            self.orch_expr,
            self.logger,
            self.stop,
            instance=instance,
        )


class InstanceAction(Action):
    def __init__(
        self,
        name,
        decorated_obj,
        pydantic_cls,
        scope=None,
        orch_expr=None,
        logger=None,
        stop=False,
        instance=None,
    ):
        super().__init__(name, decorated_obj, scope, orch_expr, logger, stop)
        self.instance = instance
        self.pydantic_cls = pydantic_cls

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self.logger:
            self.logger.debug(
                {
                    "message": f"[Action {self.name}, method {self.__name__}] Calling action: {self.name} with args: {args}"
                }
            )
        response = self.decorated_method(self.instance, *args, **kwargs)
        if self.logger:
            self.logger.debug(
                {
                    "message": f"[Action {self.name}, method {self.__name__}] Received response: {response}"
                }
            )
        return response


class ActionHandlers:
    def __init__(self, *args, **kwargs):
        self.name_to_action: Dict[str, Action] = {}

    def contains(self, name) -> bool:
        return name in self.name_to_action

    @classmethod
    def from_actions(cls, actions):
        ret = cls()
        ret.name_to_action = {action.name: action for action in actions}
        return ret

    def __len__(self) -> int:
        return len(self.name_to_action)

    def scope(self, scope):
        return {
            name: action
            for name, action in self.name_to_action.items()
            if action.scope == scope
        }

    def check_orchestration_expr_validity(self, expr):
        if expr is None:
            return

        if isinstance(expr, str):
            if expr not in self.name_to_action:
                raise ActionException(f"Action {expr} not found.")
            return

        for element in expr:
            self.check_orchestration_expr_validity(element)

    @classmethod
    def merge(cls, *handlers) -> ActionHandlers:
        merged = cls()
        for handler in handlers:
            merged.name_to_action.update(handler.name_to_action)
        return merged

    def __getitem__(self, key) -> Action:
        return self.name_to_action[key]



import types

import uuid

from pydantic import BaseModel

from actionweaver.actions.action import Action, action



def truncated_uuid4():
    return str(uuid.uuid4())[:8]


def from_model(
    model: BaseModel, stop=True, name: str = None, description: str = None
) -> Action:
    def func(*args, **kwargs):
        if args or len(kwargs) > 1:
            raise ValueError(
                f"Invalid input: The method should not have positional arguments and should only accept one keyword argument: {model.__name__.lower()} of type {model.__name__}",
                f"args: {args}",
                f"kwargs: {kwargs}",
            )

        if kwargs:
            (key,) = kwargs.keys()

            if key != model.__name__.lower():
                raise ValueError(
                    f"Invalid input: The method should accept a single keyword argument: {model.__name__.lower()}",
                    f"instead: {key}",
                )

            (value,) = kwargs.values()

            try:
                ret = model(**value)
                return ret
            except Exception as e:
                raise ValueError(
                    f"Failed to parse with the {model.__name__.lower()} model",
                    f"value: {value}",
                )


    if name is None:
        name = f"Create{model.__name__}"

    if description is None:
        description = f"Extract {model.__name__}"

    func.__doc__ = description
    func.__name__ = f"create_{model.__name__.lower()}_{truncated_uuid4()}"

    return action(name=name, stop=stop)(func).build_pydantic_model_cls(
        override_params={model.__name__.lower(): (model, ...)}
    )



from typing import List

from actionweaver.actions.action import Action, action


def repeat(
    act: Action, name: str = None, description: str = None, reducer=None
) -> Action:
    if reducer is None:
        reducer = lambda l: "\n".join([str(e) for e in l])

    def func(*args, **kwargs):
        if args or len(kwargs) > 1:
            raise ValueError(
                f"Invalid input: The method should not have positional arguments and should only accept one keyword argument",
                f"args: {args}",
                f"kwargs: {kwargs}",
            )

        if kwargs:
            (key,) = kwargs.keys()

            if key != act.name:
                raise ValueError(
                    f"Invalid input: The method should accept a single keyword argument: {act.name}",
                    f"instead: {key}",
                )

            (value,) = kwargs.values()
            if not isinstance(value, list):
                raise ValueError(
                    f"Invalid input: The {act.name} should have a list of {act.pydantic_cls.__name__} objects",
                    f"instead: {value}",
                )

            return reducer([act(**e) for e in value])

    if name is None:
        name = act.name

    if description is None:
        description = act.description

    func.__name__ = act.__name__
    func.__doc__ = description

    return action(name=name, stop=act.stop)(func).build_pydantic_model_cls(
        override_params={act.name: (List[act.pydantic_cls], ...)}
    )



import collections
import logging
import time
from typing import Dict


class TokenUsageTrackerException(Exception):
    pass


class TokenUsageTracker:
    def __init__(self, budget=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.tracker = collections.Counter()
        self.budget = budget

    def clear(self):
        self.tracker = collections.Counter()
        return self

    def track_usage(self, usage: Dict):
        self.tracker = self.tracker + collections.Counter(usage)

        self.logger.debug(
            {
                "message": "token usage updated",
                "usage": usage,
                "total_usage": dict(self.tracker),
                "timestamp": time.time(),
                "budget": self.budget,
            },
        )
        if self.budget is not None and self.tracker["total_tokens"] > self.budget:
            self.logger.error(
                {
                    "message": "Token budget exceeded",
                    "usage": usage,
                    "total_usage": dict(self.tracker),
                    "budget": self.budget,
                },
                exc_info=True,
            )
            raise TokenUsageTrackerException(
                f"Token budget exceeded. Budget: {self.budget}, Usage: {dict(self.tracker)}"
            )
        return self.tracker



from collections import UserList
from dataclasses import dataclass

from actionweaver.utils import DEFAULT_ACTION_SCOPE

#################################################################################################
# The following classes are used within action decorators
#################################################################################################


class SelectOne(UserList):
    """
    This class represents a sequence of actions. When the first action is invoked,
    it prompts for a choice from the remaining actions. For instance, if you have
    a list [action1, action2, action3], invoking action1 will prompt a choice to select
    one action from action2 and action3, or not action (default) .
    """

    def __eq__(self, other):
        return type(self) == type(other) and self.data == other.data


class RequireNext(UserList):
    """
    This class represents a sequence of actions. Invoking the first action prompts
    the llm to proceed with the next action in the sequence. For example, if you have
    [action1, action2, action3], invoking action1 will prompt the llm to continue
    with action2 and subsequently action3.
    """

    def __eq__(self, other):
        return type(self) == type(other) and self.data == other.data


#################################################################################################
# The following classes are used within action handlers class, users should not use them directly
#################################################################################################


@dataclass
class _ActionHandlerLLMInvoke:
    scope: str = DEFAULT_ACTION_SCOPE

    def __hash__(self):
        return hash(f"_ActionHandlerLLMInvoke[scope={self.scope}]")


@dataclass
class _ActionHandlerRequired:
    action: str


class _ActionHandlerSelectOne(UserList):
    pass


class _ActionDefault:
    """This class represents the default action"""

    pass



from __future__ import annotations

import inspect
import json
import logging
import time
import uuid
from argparse import Action
from itertools import chain
from typing import List

from openai import AzureOpenAI, Stream
from openai.types.chat.chat_completion_message import (
    ChatCompletionMessage,
    FunctionCall,
)

from actionweaver.actions.action import ActionHandlers
from actionweaver.actions.orchestration import (
    build_orchestration_dict,
    parse_orchestration_expr,
)
from actionweaver.actions.orchestration_expr import (
    _ActionDefault,
    _ActionHandlerLLMInvoke,
)
from actionweaver.llms.azure.functions import Functions
from actionweaver.llms.azure.tokens import TokenUsageTracker
from actionweaver.utils import DEFAULT_ACTION_SCOPE
from actionweaver.utils.stream import get_first_element_and_iterator, merge_dicts

# TODO: support AsyncAzureOpenAI


class ChatCompletionException(Exception):
    pass


class ChatCompletion:
    def __init__(
        self,
        model,
        azure_endpoint,
        api_key,
        api_version,
        token_usage_tracker=None,
        logger=None,
        azure_deployment="",
    ):
        self.model = model
        self.action_handlers = ActionHandlers()
        self.logger = logger or logging.getLogger(__name__)
        if azure_deployment == "":
            self.client = AzureOpenAI(
                azure_endpoint=azure_endpoint,
                api_key=api_key,
                api_version=api_version
            )
        else:
            self.client = AzureOpenAI(
                azure_endpoint=azure_endpoint,
                api_key=api_key,
                api_version=api_version,
                azure_deployment=azure_deployment
            )
        self.token_usage_tracker = token_usage_tracker or TokenUsageTracker(
            logger=logger
        )

    def _bind_action_handlers(self, action_handlers: ActionHandlers) -> ChatCompletion:
        self.action_handlers = action_handlers
        return self

    def _invoke_function(
        self,
        call_id,
        messages,
        model,
        function_call,
        orchestration_dict,
        default_expr,
        action_handler: ActionHandlers,
        functions,
    ):
        """Invoke the function, update the messages, returns functions argument for the next OpenAI API call or halt the function loop and return the response."""

        if isinstance(function_call, FunctionCall):
            function_call = function_call.model_dump()

        messages += [
            {
                "role": "assistant",
                "content": None,
                "function_call": function_call,
            }
        ]

        name = function_call["name"]
        arguments = function_call["arguments"]

        if action_handler.contains(name):
            try:
                arguments = json.loads(function_call["arguments"])
            except json.decoder.JSONDecodeError as e:
                self.logger.error(
                    {
                        "message": "Parsing function call arguments from OpenAI response ",
                        "arguments": function_call["arguments"],
                        "timestamp": time.time(),
                        "model": model,
                        "call_id": call_id,
                    },
                    exc_info=True,
                )
                raise ChatCompletionException(e) from e

            # Invoke action
            function_response = action_handler[name](**arguments)
            stop = action_handler[name].stop
            messages += [
                {
                    "role": "function",
                    "name": name,
                    "content": str(function_response),
                }
            ]

            self.logger.debug(
                {
                    "message": "Action invoked and response received",
                    "action_name": name,
                    "action_arguments": arguments,
                    "action_response": function_response,
                    "timestamp": time.time(),
                    "call_id": call_id,
                    "stop": stop,
                }
            )

            # Update new functions for next OpenAI api call
            # if name doesn't exist in orchestration dict, use _ActionDefaultLLM which doesn't invoke functions
            expr = (
                orchestration_dict[name]
                if name in orchestration_dict
                else _ActionDefault()
            )
            return (
                Functions.from_expr(
                    expr,
                    action_handler,
                ),
                (stop, function_response),
            )
        else:
            unavailable_function_msg = f"{name} is not a valid function name, use one of the following: {', '.join([func['name'] for func in functions.functions])}"
            messages += [
                {
                    "role": "user",
                    "content": unavailable_function_msg,
                }
            ]
            self.logger.debug(
                {
                    "message": "Unavailable action",
                    "action_name": name,
                    "action_arguments": arguments,
                    "action_response": unavailable_function_msg,
                    "timestamp": time.time(),
                    "call_id": call_id,
                }
            )
            return functions, (False, None)

    def create(
        self,
        messages,
        *args,
        scope=None,
        orch_expr=None,
        stream=False,
        action_handler: ActionHandlers = None,
        actions: List[Action] = [],
        **kwargs,
    ):
        """
        Invoke the OpenAI API with the provided messages and functions.

        Functions supplied to the API are determined by the specified scope and orchestration expression.

        Args:
            messages (list): List of message objects for the conversation.
            scope (str): Scope of the functions to be used.
            orch_expr: If specified, overrides the orchestration dictionary, determining the API call flow.
            stream (bool): If True, returns a generator that yields the API responses.

        Returns:
            API response with generated output.
        """

        # Todo: pass call_id to the decorated method
        call_id = str(uuid.uuid4())

        model = kwargs.get("model", self.model)
        temperature = kwargs.get("temperature", 0.0)

        # Restart token usage tracker
        self.token_usage_tracker.clear()

        if scope is None:
            scope = DEFAULT_ACTION_SCOPE
        default_expr = _ActionHandlerLLMInvoke(scope)

        # TODO: move below into init method
        # initialize action handlers
        if action_handler is None:
            if actions:
                action_handler = ActionHandlers.from_actions(actions)
            else:
                action_handler = self.action_handlers

        for _, action in action_handler.name_to_action.items():
            action_handler.check_orchestration_expr_validity(action.orch_expr)

        # Build orchestration
        orchestration_dict = build_orchestration_dict(action_handler)

        # If an orchestration expression is provided, override the orchestration dictionary
        if orch_expr is not None:
            action_handler.check_orchestration_expr_validity(orch_expr)

            scope = "_llm_orchestration"
            # Construct a new orchestration dictionary using the parsed orchestration expression
            new_orchestration_dict = parse_orchestration_expr([scope] + orch_expr)

            # Replace the default expression with the new one
            default_expr = _ActionHandlerLLMInvoke(scope)
            new_orchestration_dict[default_expr] = new_orchestration_dict.pop(scope)

            # Update the orchestration dictionary with the new one
            orchestration_dict = new_orchestration_dict

        self.logger.debug(
            {
                "message": "Creating new chat completion",
                "model": model,
                "scope": scope,
                "input_messages": messages,
                "timestamp": time.time(),
                "call_id": call_id,
            }
        )

        response = None
        functions_to_be_called_with = Functions.from_expr(
            orchestration_dict[default_expr],
            action_handler,
        )

        while True:
            self.logger.debug(
                {
                    "message": "Calling OpenAI API",
                    "call_id": call_id,
                    "input_messages": messages,
                    "model": model,
                    "scope": scope,
                    "timestamp": time.time(),
                    **functions_to_be_called_with.to_arguments(),
                }
            )

            function_argument = functions_to_be_called_with.to_arguments()

            if function_argument["functions"]:
                api_response = self.client.chat.completions.create(
                    model=model,
                    temperature=temperature,
                    messages=messages,
                    stream=stream,
                    **function_argument,
                )
            else:
                api_response = self.client.chat.completions.create(
                    model=model,
                    temperature=temperature,
                    messages=messages,
                    stream=stream,
                )

            self.logger.debug(
                {
                    "message": "Received response from OpenAI API",
                    "response": api_response,
                    "timestamp": time.time(),
                    "call_id": call_id,
                }
            )

            # logic to handle streaming API response
            stream_function_calls = None
            if isinstance(api_response, Stream):
                _, iterator = get_first_element_and_iterator(api_response)

                # TODO: move this to open ai as well

                for chunk in iterator:
                    if chunk.choices:
                        delta = chunk.choices[0].delta

                        if delta.function_call:
                            if stream_function_calls is None:
                                # if function call detected, we merge all deltas and treat it as the non-stream response
                                stream_function_calls = FunctionCall(
                                    name="", arguments=""
                                )

                            stream_function_calls.name += (
                                delta.function_call.name
                                if delta.function_call.name
                                else ""
                            )

                            stream_function_calls.arguments += (
                                delta.function_call.arguments
                                if delta.function_call.arguments
                                else ""
                            )

                        elif delta.content:
                            # if it has content return as generator right away
                            return chain([chunk], iterator)
                        else:
                            self.logger.debug(
                                {
                                    "message": "Unsupported streaming response",
                                    "chunk": str(chunk),
                                    "timestamp": time.time(),
                                    "call_id": call_id,
                                }
                            )

            else:
                self.token_usage_tracker.track_usage(api_response.usage)

            if stream_function_calls is not None:
                functions_to_be_called_with, (stop, resp) = self._invoke_function(
                    call_id,
                    messages,
                    model,
                    stream_function_calls,
                    orchestration_dict,
                    default_expr,
                    action_handler,
                    functions_to_be_called_with,
                )
                if stop:
                    return resp
            else:
                choice = api_response.choices[0]
                message = choice.message

                if message.function_call:
                    functions_to_be_called_with, (stop, resp) = self._invoke_function(
                        call_id,
                        messages,
                        model,
                        message.function_call,
                        orchestration_dict,
                        default_expr,
                        action_handler,
                        functions_to_be_called_with,
                    )
                    if stop:
                        return resp
                elif message.content is not None:
                    response = message.content

                    # ignore last message in the function loop
                    # messages += [{"role": "assistant", "content": message["content"]}]
                    if choice.finish_reason == "stop":
                        """
                        Stop Reasons:

                        - Occurs when the API returns a message that is complete or is concluded by one of the stop sequences defined via the 'stop' parameter.

                        See https://platform.openai.com/docs/guides/gpt/chat-completions-api for details.
                        """
                        self.logger.debug(
                            {
                                "message": "Model decides to stop",
                                "model": model,
                                "timestamp": time.time(),
                                "call_id": call_id,
                            }
                        )

                        break
                else:
                    raise ChatCompletionException(
                        f"Unsupported response from OpenAI api: {api_response}"
                    )
        return response



from actionweaver.utils import DEFAULT_ACTION_SCOPE

from .orchestration_expr import (
    RequireNext,
    SelectOne,
    _ActionDefault,
    _ActionHandlerLLMInvoke,
    _ActionHandlerRequired,
    _ActionHandlerSelectOne,
)


class ActionOrchestrationParseError(Exception):
    pass


class Orchestration:
    def __init__(self, data: dict = None):
        self.data = {}

        if data is not None:
            self.data |= data

    def pop(self, key, default=None):
        """
        Remove the item with the specified key from the data dictionary and return its value.
        If the key is not found and a default value is provided, return the default value.
        If the key is not found and no default value is provided, raise a KeyError.
        """
        if key in self.data:
            return self.data.pop(key)
        elif default is not None:
            return default
        else:
            raise KeyError(key)

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def keys(self):
        return list(self.data.keys())

    def values(self):
        return list(self.data.values())

    def items(self):
        return list(self.data.items())

    def get(self, key, default=None):
        return self.data.get(key, default)

    def clear(self):
        self.data.clear()

    def update(self, *args, **kwargs):
        self.data.update(*args, **kwargs)

    def __contains__(self, key):
        return key in self.data

    def __len__(self):
        return len(self.data)

    def __str__(self):
        return str(self.data)

    def __eq__(self, other):
        if isinstance(other, Orchestration):
            return self.data == other.data
        return False

    def to_pyvis_network(self):
        """Experimental for debugging purpose"""
        from pyvis.network import Network

        def _generate_node_config(n):
            if isinstance(n, _ActionHandlerLLMInvoke):
                if n.scope == DEFAULT_ACTION_SCOPE:
                    return {
                        "n_id": str(n),
                        "label": "LLM[DEFAULT]",
                        "color": "#005B00",
                        "physics": False,
                    }
                else:
                    return {
                        "n_id": str(n),
                        "label": f"LLM[{n.scope}]",
                        "color": "#00CC99",
                        "physics": False,
                    }
            elif isinstance(n, str):
                return {
                    "n_id": n,
                    "label": None,
                    "color": "#FF7769",
                    "physics": False,
                }
            else:
                raise ActionException("unsupported node type for visualization")

        g = Network(notebook=True, directed=True)

        for node, expr in self.orch_dict.items():
            g.add_node(**_generate_node_config(node))

            if isinstance(expr, _ActionHandlerSelectOne):
                for dest in expr:
                    g.add_node(**_generate_node_config(dest))
                    g.add_edge(
                        _generate_node_config(node)["n_id"],
                        _generate_node_config(dest)["n_id"],
                        color="#00CC99",
                        label="select",
                        physics=False,
                    ),
            elif isinstance(expr, _ActionHandlerRequired):
                g.add_node(**_generate_node_config(expr.action))
                g.add_edge(
                    _generate_node_config(node)["n_id"],
                    _generate_node_config(expr.action)["n_id"],
                    color="#FF0000",
                    label="requires",
                    physics=False,
                ),
        return g


def parse_orchestration_expr(expr):
    orch_dict = Orchestration()

    def get_first_action(l):
        """
        Get first action name from list expression.
        """
        if isinstance(l, (SelectOne, RequireNext)):
            return get_first_action(l[0])
        elif isinstance(l, str):
            return l
        else:
            raise ActionOrchestrationParseError(
                f"Invalid object in orchestration: {l}."
            )

    def get_last_action(l):
        """
        Get last action name from list expression.
        """
        if isinstance(l, SelectOne):
            raise ActionOrchestrationParseError(
                f"Can't decide last action of SelectOne: {l}."
            )
        elif isinstance(l, RequireNext):
            return get_last_action(l[-1])
        elif isinstance(l, str):
            return l
        else:
            raise ActionOrchestrationParseError(
                f"Invalid object in orchestration: {l}."
            )

    def parse(l):
        """
        Parse list expression into orch_dict.
        """

        if isinstance(l, SelectOne):
            curr = get_last_action(l[0])
            parse(l[0])

            next_actions = []
            for n in list(l)[1:]:
                next_actions.append(get_first_action(n))
                parse(n)

            if curr in orch_dict:
                raise ActionOrchestrationParseError(f"Inconsistency caused by {l}")
            orch_dict[curr] = _ActionHandlerSelectOne(next_actions)
        elif isinstance(l, RequireNext):
            parse(l[0])
            prev, curr = get_last_action(l[0]), None
            for n in range(1, len(l)):
                curr = get_first_action(l[n])

                if prev in orch_dict:
                    raise ActionOrchestrationParseError(f"Inconsistency caused by {l}")
                orch_dict[prev] = _ActionHandlerRequired(get_first_action(curr))
                parse(l[n])

                prev = curr
        elif isinstance(l, str):
            return
        else:
            raise ActionOrchestrationParseError(f"Invalid orchestration expr: {l}.")

    parse(expr)
    return orch_dict


# TODO: solve circular import from importing ActionHandlers
def build_orchestration_dict(action_handlers) -> Orchestration:
    """
    Parse orchestration expressions from all actions.
    """
    # orch_dict = {
    #     _ActionHandlerLLMInvoke(DEFAULT_ACTION_SCOPE): _ActionHandlerSelectOne([])
    # }

    orch_dict = Orchestration()
    orch_dict[_ActionHandlerLLMInvoke(DEFAULT_ACTION_SCOPE)] = _ActionHandlerSelectOne(
        []
    )

    for _, action in action_handlers.name_to_action.items():
        if action.orch_expr:
            if action.orch_expr[0] != action.name:
                raise ActionOrchestrationParseError(
                    f"Invalid expression {action.orch_expr} in Action {action.name}, "
                    "The first element of the orchestration expression must be the action name. "
                    "For example, in 'action1', 'SelectOne' should be used as 'SelectOne([action1, ...])', "
                    "and 'RequireNext' should be used as 'RequireNext([action1, ...])'."
                )

            orch_dict.update(**parse_orchestration_expr(action.orch_expr))

        # Parse scope
        if (
            _ActionHandlerLLMInvoke(action.scope) not in orch_dict
            or orch_dict[_ActionHandlerLLMInvoke(action.scope)] is None
        ):  # if scope doesn't exist or map to None like DEFAULT_ACTION_SCOPE, then create a new one
            orch_dict[_ActionHandlerLLMInvoke(action.scope)] = _ActionHandlerSelectOne(
                []
            )
        elif not isinstance(
            orch_dict[_ActionHandlerLLMInvoke(action.scope)],
            _ActionHandlerSelectOne,
        ):
            raise ActionOrchestrationParseError(
                f"The scope {action.scope} of Action {action.name} causes inconsistency in orchestration."
            )
        orch_dict[_ActionHandlerLLMInvoke(action.scope)].append(action.name)

    return orch_dict



from inspect import getfullargspec
from typing import Any, Callable, Type

from pydantic import BaseModel, create_model


def create_pydantic_model_from_func(
    func: Callable,
    model_name: str,
    base_model: Type[BaseModel] = BaseModel,
    models=None,  # models: Optional pydantic models needed for the pydantic model from function signature
    override_params=None,  # override_params: Optional dictionary of parameters to override kwarg and non-kwarg.
):
    if models is None:
        models = []

    # Retrieve function signature details
    """
    - args is a list of the positional parameter names.
    - varargs is the name of the * parameter or None if arbitrary positional arguments are not accepted.
    - varkw is the name of the ** parameter or None if arbitrary keyword arguments are not accepted.
    - defaults is an n-tuple of default argument values corresponding to the last n positional parameters, or None if there are no such defaults defined.
    - kwonlyargs is a list of keyword-only parameter names in declaration order.
    - kwonlydefaults is a dictionary mapping parameter names from kwonlyargs to the default values used if no argument is supplied.
    - annotations is a dictionary mapping parameter names to annotations.
    """
    (
        args,
        varargs,
        varkw,
        defaults,
        kwonlyargs,
        kwonlydefaults,
        annotations,
    ) = getfullargspec(func)

    if len(args) > 0 and args[0] == "self":
        args = args[1:]  # skip self

    defaults = defaults or []
    args = args or []

    # Calculate the number of arguments that don't have default values
    non_default_args = len(args) - len(defaults)

    # Add placeholders for non-default arguments
    defaults = (...,) * non_default_args + tuple(defaults)

    # Create a dictionary for keyword-only arguments with their defaults
    keyword_only_params = {
        param: kwonlydefaults.get(param, Any) for param in kwonlyargs
    }

    """ Create a dictionary for all arguments with their annotations and defaults
    e.g. 
    {
        'name': ('str', 'Default Name'),
        'age': ('int', 0),
        'email': ('str', None),
    }
    """

    params = {}

    if override_params is None:
        models_dict = {model.__name__: model for model in models}
        for param, default in zip(args, defaults):
            annotation = annotations.get(param, Any)
            annotation = models_dict.get(annotation, annotation)
            params[param] = (annotation, default)
    else:
        # use override_params instead
        params = override_params
        keyword_only_params = {}

    # Configure the class to allow extra parameters if **kwargs is in the function signature
    class Config:
        extra = "allow"

    config = Config if varkw else None

    # Create and return the pydantic model using the gathered parameters, configurations, and base model
    return create_model(
        model_name,
        **params,
        **keyword_only_params,
        __base__=base_model,
        # __config__=config,
    )



from openai import OpenAI

client = OpenAI()

from actionweaver import action


class OpenAIAPI:
    @action(name="GenerateImage")
    def generate_image(self, prompt: str) -> str:
        """
        Generates an image based on a given prompt.

        Args:
            prompt (str): A descriptive text input that guides the content of the generated image (e.g., "lion running on Mars").

        Returns:
            str: A message indicating that the image has been created, along with the URL of the generated image.
        """
        from IPython.display import Image, display

        response = client.images.generate(prompt=prompt, n=1, size="1024x1024")

        url = response.data[0].url
        display(Image(url=url))
        return "The image has been created"



# These modules provide mixin examples that are aimed at boosting agent development speed.
# They also offer a foundation for creating your own custom mixins.

from .folium import Folium
from .langchain import LangChainTools
from .openai import OpenAIAPI



from actionweaver.actions.orchestration_expr import (
    _ActionDefault,
    _ActionHandlerLLMInvoke,
    _ActionHandlerRequired,
    _ActionHandlerSelectOne,
)


class FunctionException(Exception):
    pass


class Functions:
    def __init__(self, function_call=None, functions=None) -> None:
        self.functions = functions
        self.function_call = function_call

    @classmethod
    def from_expr(cls, expr, action_handlers):
        if isinstance(expr, (_ActionHandlerLLMInvoke, _ActionDefault)):
            return cls()
        elif isinstance(expr, _ActionHandlerRequired):
            return cls(
                function_call={"name": expr.action},
                functions=[
                    {
                        "name": action_handlers.name_to_action[expr.action].name,
                        "description": action_handlers.name_to_action[
                            expr.action
                        ].description,
                        "parameters": action_handlers.name_to_action[
                            expr.action
                        ].json_schema(),
                    }
                ],
            )
        elif isinstance(expr, _ActionHandlerSelectOne):
            return cls(
                functions=[
                    {
                        "name": action_handlers.name_to_action[action].name,
                        "description": action_handlers.name_to_action[
                            action
                        ].description,
                        "parameters": action_handlers.name_to_action[
                            action
                        ].json_schema(),
                    }
                    for action in expr
                ],
                function_call="auto",
            )
        else:
            raise FunctionException(f"Invalid orchestration expression: {expr}")

    def to_arguments(self):
        return {
            "functions": self.functions,
            "function_call": self.function_call,
        }
